@@ -4,6 +4,8 @@ import hashlib
 import requests
 import time
 import json
+import cgi
+from StringIO import StringIO
 
 from xml.dom import minidom
 
@@ -19,7 +21,8 @@ class WechatBasic(object):
     仅包含官方 API 中所包含的内容, 如需高级功能支持请移步 ext.py 中的 WechatExt 类
     """
     def __init__(self, token=None, appid=None, appsecret=None, partnerid=None,
-                 partnerkey=None, paysignkey=None, access_token=None, access_token_expires_at=None):
+                 partnerkey=None, paysignkey=None, access_token=None, access_token_expires_at=None,
+                 jsapi_ticket=None, jsapi_ticket_expires_at=None):
         """
         :param token: 微信 Token
         :param appid: App ID
@@ -29,6 +32,8 @@ class WechatBasic(object):
         :param paysignkey: 商户签名密钥 Key, 支付权限专用
         :param access_token: 直接导入的 access_token 值, 该值需要在上一次该类实例化之后手动进行缓存并在此处传入, 如果不传入, 将会在需要时自动重新获取
         :param access_token_expires_at: 直接导入的 access_token 的过期日期，该值需要在上一次该类实例化之后手动进行缓存并在此处传入, 如果不传入, 将会在需要时自动重新获取
+        :param jsapi_ticket: 直接导入的 jsapi_ticket 值, 该值需要在上一次该类实例化之后手动进行缓存并在此处传入, 如果不传入, 将会在需要时自动重新获取
+        :param jsapi_ticket_expires_at: 直接导入的 jsapi_ticket 的过期日期，该值需要在上一次该类实例化之后手动进行缓存并在此处传入, 如果不传入, 将会在需要时自动重新获取
         """
         self.__token = token
         self.__appid = appid
@@ -39,6 +44,8 @@ class WechatBasic(object):
 
         self.__access_token = access_token
         self.__access_token_expires_at = access_token_expires_at
+        self.__jsapi_ticket = jsapi_ticket
+        self.__jsapi_ticket_expires_at = jsapi_ticket_expires_at
         self.__is_parse = False
         self.__message = None
 
@@ -58,11 +65,34 @@ class WechatBasic(object):
         tmp_list = [self.__token, timestamp, nonce]
         tmp_list.sort()
         tmp_str = ''.join(tmp_list)
-        if signature == hashlib.sha1(tmp_str).hexdigest():
+        if signature == hashlib.sha1(tmp_str.encode('utf-8')).hexdigest():
             return True
         else:
             return False
 
+    def generate_jsapi_signature(self, timestamp, noncestr, url, jsapi_ticket=None):
+        """
+        使用 jsapi_ticket 对 url 进行签名
+        :param timestamp: 时间戳
+        :param noncestr: 随机数
+        :param url: 要签名的 url，不包含 # 及其后面部分
+        :param jsapi_ticket: (可选参数) jsapi_ticket 值 (如不提供将自动通过 appid 和 appsecret 获取)
+        :return: 返回sha1签名的hexdigest值
+        """
+        if not jsapi_ticket:
+            jsapi_ticket = self.jsapi_ticket
+        data = {
+            'jsapi_ticket': jsapi_ticket,
+            'noncestr': noncestr,
+            'timestamp': timestamp,
+            'url': url,
+        }
+        keys = data.keys()
+        keys.sort()
+        data_str = '&'.join(['%s=%s' % (key, data[key]) for key in keys])
+        signature = hashlib.sha1(data_str.encode('utf-8')).hexdigest()
+        return signature
+
     def parse_data(self, data):
         """
         解析微信服务器发送过来的数据并保存类中
@@ -117,14 +147,29 @@ class WechatBasic(object):
             'access_token_expires_at': self.__access_token_expires_at,
         }
 
-    def response_text(self, content):
+    def get_jsapi_ticket(self):
+        """
+        获取 Jsapi Ticket 及 Jsapi Ticket 过期日期, 仅供缓存使用, 如果希望得到原生的 Jsapi Ticket 请求数据请使用 :func:`grant_jsapi_ticket`
+        :return: dict 对象, key 包括 `jsapi_ticket` 及 `jsapi_ticket_expires_at`
+        """
+        self._check_appid_appsecret()
+
+        return {
+            'jsapi_ticket': self.jsapi_ticket,
+            'jsapi_ticket_expires_at': self.__jsapi_ticket_expires_at,
+        }
+
+    def response_text(self, content, escape=False):
         """
         将文字信息 content 组装为符合微信服务器要求的响应数据
         :param content: 回复文字
+        :param escape: 是否转义该文本内容 (默认不转义)
         :return: 符合微信服务器要求的 XML 响应数据
         """
         self._check_parse()
         content = self._transcoding(content)
+        if escape:
+            content = cgi.escape(content)
 
         return TextReply(message=self.__message, content=content).render()
 
@@ -204,15 +249,17 @@ class WechatBasic(object):
             news.add_article(article)
         return news.render()
 
-    def grant_token(self):
+    def grant_token(self, override=True):
         """
         获取 Access Token
         详情请参考 http://mp.weixin.qq.com/wiki/11/0e4b294685f817b95cbed85ba5e82b8f.html
+        :param override: 是否在获取的同时覆盖已有 access_token (默认为True)
         :return: 返回的 JSON 数据包
+        :raise HTTPError: 微信api http 请求失败
         """
         self._check_appid_appsecret()
 
-        return self._get(
+        response_json = self._get(
             url="https://api.weixin.qq.com/cgi-bin/token",
             params={
                 "grant_type": "client_credential",
@@ -220,6 +267,27 @@ class WechatBasic(object):
                 "secret": self.__appsecret,
             }
         )
+        if override:
+            self.__access_token = response_json['access_token']
+            self.__access_token_expires_at = int(time.time()) + response_json['expires_in']
+        return response_json
+
+    def grant_jsapi_ticket(self):
+        """
+        获取 Jsapi Ticket
+        详情请参考 http://mp.weixin.qq.com/wiki/7/aaa137b55fb2e0456bf8dd9148dd613f.html#.E9.99.84.E5.BD.951-JS-SDK.E4.BD.BF.E7.94.A8.E6.9D.83.E9.99.90.E7.AD.BE.E5.90.8D.E7.AE.97.E6.B3.95
+        :return: 返回的 JSON 数据包
+        :raise HTTPError: 微信api http 请求失败
+        """
+        self._check_appid_appsecret()
+
+        return self._get(
+            url="https://api.weixin.qq.com/cgi-bin/ticket/getticket",
+            params={
+                "access_token": self.access_token,
+                "type": "jsapi",
+            }
+        )
 
     def create_menu(self, menu_data):
         """
@@ -263,6 +331,7 @@ class WechatBasic(object):
         详情请参考 http://mp.weixin.qq.com/wiki/13/43de8269be54a0a6f64413e4dfa94f39.html
         :param menu_data: Python 字典
         :return: 返回的 JSON 数据包
+        :raise HTTPError: 微信api http 请求失败
         """
         self._check_appid_appsecret()
 
@@ -276,6 +345,7 @@ class WechatBasic(object):
         查询自定义菜单
         详情请参考 http://mp.weixin.qq.com/wiki/16/ff9b7b85220e1396ffa16794a9d95adc.html
         :return: 返回的 JSON 数据包
+        :raise HTTPError: 微信api http 请求失败
         """
         self._check_appid_appsecret()
 
@@ -286,20 +356,43 @@ class WechatBasic(object):
         删除自定义菜单
         详情请参考 http://mp.weixin.qq.com/wiki/16/8ed41ba931e4845844ad6d1eeb8060c8.html
         :return: 返回的 JSON 数据包
+        :raise HTTPError: 微信api http 请求失败
         """
         self._check_appid_appsecret()
 
         return self._get('https://api.weixin.qq.com/cgi-bin/menu/delete')
 
-    def upload_media(self, media_type, media_file):
+    def upload_media(self, media_type, media_file, extension=''):
         """
         上传多媒体文件
         详情请参考 http://mp.weixin.qq.com/wiki/10/78b15308b053286e2a66b33f0f0f5fb6.html
         :param media_type: 媒体文件类型，分别有图片（image）、语音（voice）、视频（video）和缩略图（thumb）
-        :param media_file:要上传的文件，一个 File-object
+        :param media_file: 要上传的文件，一个 File object 或 StringIO object
+        :param extension: 如果 media_file 传入的为 StringIO object，那么必须传入 extension 显示指明该媒体文件扩展名，如 ``mp3``, ``amr``；如果 media_file 传入的为 File object，那么该参数请留空
         :return: 返回的 JSON 数据包
+        :raise HTTPError: 微信api http 请求失败
         """
         self._check_appid_appsecret()
+        if not isinstance(media_file, file) and not isinstance(media_file, StringIO):
+            raise ValueError('Parameter media_file must be file object or StringIO.StringIO object.')
+        if isinstance(media_file, StringIO) and extension.lower() not in ['jpg', 'jpeg', 'amr', 'mp3', 'mp4']:
+            raise ValueError('Please provide \'extension\' parameters when the type of \'media_file\' is \'StringIO.StringIO\'.')
+        if isinstance(media_file, file):
+            extension = media_file.name.split('.')[-1]
+            if extension.lower() not in ['jpg', 'jpeg', 'amr', 'mp3', 'mp4']:
+                raise ValueError('Invalid file type.')
+
+        ext = {
+            'jpg': 'image/jpeg',
+            'jpeg': 'image/jpeg',
+            'amr': 'audio/amr',
+            'mp3': 'audio/mpeg',
+            'mp4': 'video/mp4',
+        }
+        if isinstance(media_file, StringIO):
+            filename = 'temp.' + extension
+        else:
+            filename = media_file.name
 
         return self._post(
             url='http://file.api.weixin.qq.com/cgi-bin/media/upload',
@@ -308,7 +401,7 @@ class WechatBasic(object):
                 'type': media_type,
             },
             files={
-                'media': media_file,
+                'media': (filename, media_file, ext[extension])
             }
         )
 
@@ -336,6 +429,7 @@ class WechatBasic(object):
         详情请参考 http://mp.weixin.qq.com/wiki/13/be5272dc4930300ba561d927aead2569.html
         :param name: 分组名字（30个字符以内）
         :return: 返回的 JSON 数据包
+        :raise HTTPError: 微信api http 请求失败
         """
         self._check_appid_appsecret()
 
@@ -353,6 +447,7 @@ class WechatBasic(object):
         查询所有分组
         详情请参考 http://mp.weixin.qq.com/wiki/13/be5272dc4930300ba561d927aead2569.html
         :return: 返回的 JSON 数据包
+        :raise HTTPError: 微信api http 请求失败
         """
         self._check_appid_appsecret()
 
@@ -364,6 +459,7 @@ class WechatBasic(object):
         详情请参考 http://mp.weixin.qq.com/wiki/13/be5272dc4930300ba561d927aead2569.html
         :param openid: 用户的OpenID
         :return: 返回的 JSON 数据包
+        :raise HTTPError: 微信api http 请求失败
         """
         self._check_appid_appsecret()
 
@@ -381,6 +477,7 @@ class WechatBasic(object):
         :param group_id: 分组id，由微信分配
         :param name: 分组名字（30个字符以内）
         :return: 返回的 JSON 数据包
+        :raise HTTPError: 微信api http 请求失败
         """
         self._check_appid_appsecret()
 
@@ -401,6 +498,7 @@ class WechatBasic(object):
         :param user_id: 用户 ID 。 就是你收到的 WechatMessage 的 source
         :param group_id: 分组 ID
         :return: 返回的 JSON 数据包
+        :raise HTTPError: 微信api http 请求失败
         """
         self._check_appid_appsecret()
 
@@ -419,6 +517,7 @@ class WechatBasic(object):
         :param user_id: 用户 ID, 就是你收到的 WechatMessage 的 source
         :param lang: 返回国家地区语言版本，zh_CN 简体，zh_TW 繁体，en 英语
         :return: 返回的 JSON 数据包
+        :raise HTTPError: 微信api http 请求失败
         """
         self._check_appid_appsecret()
 
@@ -437,6 +536,7 @@ class WechatBasic(object):
         详情请参考 http://mp.weixin.qq.com/wiki/3/17e6919a39c1c53555185907acf70093.html
         :param first_user_id: 可选。第一个拉取的OPENID，不填默认从头开始拉取
         :return: 返回的 JSON 数据包
+        :raise HTTPError: 微信api http 请求失败
         """
         self._check_appid_appsecret()
 
@@ -454,6 +554,7 @@ class WechatBasic(object):
         :param user_id: 用户 ID, 就是你收到的 WechatMessage 的 source
         :param content: 消息正文
         :return: 返回的 JSON 数据包
+        :raise HTTPError: 微信api http 请求失败
         """
         self._check_appid_appsecret()
 
@@ -475,6 +576,7 @@ class WechatBasic(object):
         :param user_id: 用户 ID, 就是你收到的 WechatMessage 的 source
         :param media_id: 图片的媒体ID。 可以通过 :func:`upload_media` 上传。
         :return: 返回的 JSON 数据包
+        :raise HTTPError: 微信api http 请求失败
         """
         self._check_appid_appsecret()
 
@@ -496,6 +598,7 @@ class WechatBasic(object):
         :param user_id: 用户 ID, 就是你收到的 WechatMessage 的 source
         :param media_id: 发送的语音的媒体ID。 可以通过 :func:`upload_media` 上传。
         :return: 返回的 JSON 数据包
+        :raise HTTPError: 微信api http 请求失败
         """
         self._check_appid_appsecret()
 
@@ -519,6 +622,7 @@ class WechatBasic(object):
         :param title: 视频消息的标题
         :param description: 视频消息的描述
         :return: 返回的 JSON 数据包
+        :raise HTTPError: 微信api http 请求失败
         """
         self._check_appid_appsecret()
 
@@ -550,6 +654,7 @@ class WechatBasic(object):
         :param title: 音乐标题
         :param description: 音乐描述
         :return: 返回的 JSON 数据包
+        :raise HTTPError: 微信api http 请求失败
         """
         self._check_appid_appsecret()
 
@@ -608,6 +713,7 @@ class WechatBasic(object):
         详情请参考 http://mp.weixin.qq.com/wiki/18/28fc21e7ed87bec960651f0ce873ef8a.html
         :param data: 你要发送的参数 dict
         :return: 返回的 JSON 数据包
+        :raise HTTPError: 微信api http 请求失败
         """
         self._check_appid_appsecret()
 
@@ -640,11 +746,22 @@ class WechatBasic(object):
             now = time.time()
             if self.__access_token_expires_at - now > 60:
                 return self.__access_token
-        response_json = self.grant_token()
-        self.__access_token = response_json['access_token']
-        self.__access_token_expires_at = int(time.time()) + response_json['expires_in']
+        self.grant_token()
         return self.__access_token
 
+    @property
+    def jsapi_ticket(self):
+        self._check_appid_appsecret()
+
+        if self.__jsapi_ticket:
+            now = time.time()
+            if self.__jsapi_ticket_expires_at - now > 60:
+                return self.__jsapi_ticket
+        response_json = self.grant_jsapi_ticket()
+        self.__jsapi_ticket = response_json['ticket']
+        self.__jsapi_ticket_expires_at = int(time.time()) + response_json['expires_in']
+        return self.__jsapi_ticket
+
     def _check_token(self):
         """
         检查 Token 是否存在
@@ -684,6 +801,7 @@ class WechatBasic(object):
         :param url: 请求地址
         :param kwargs: 附加数据
         :return: 微信服务器响应的 json 数据
+        :raise HTTPError: 微信api http 请求失败
         """
         if "params" not in kwargs:
             kwargs["params"] = {
@@ -710,6 +828,7 @@ class WechatBasic(object):
         :param url: 请求地址
         :param kwargs: 附加数据
         :return: 微信服务器响应的 json 数据
+        :raise HTTPError: 微信api http 请求失败
         """
         return self._request(
             method="get",
@@ -723,6 +842,7 @@ class WechatBasic(object):
         :param url: 请求地址
         :param kwargs: 附加数据
         :return: 微信服务器响应的 json 数据
+        :raise HTTPError: 微信api http 请求失败
         """
         return self._request(
             method="post",
@@ -746,4 +866,4 @@ class WechatBasic(object):
             result = data.decode('utf-8')
         else:
             raise ParseError()
-        return result
\ No newline at end of file
+        return result
